/*********************************************************************
    This file is part of Siworin. Copyright © 2021 by Ruud Harmsen.

    Siworin is a simple word indexer and local search engine. See
    https://rudhar.com/sfreview/siworin/#siw04 and
    https://rudhar.com/sfreview/siworin/toolsrc/siworin .

    Siworin is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published
    by the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    Siworin is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Siworin.  If not, see <https://www.gnu.org/licenses/>.
  ********************************************************************/

/* 5 and 6 November 2022:
   Fix a bug. For example when three words matched, adding
   an extra search word, occurring in the same file but too
   far away, made the whole match disappear because of the
   too great distance. But of course even though the 4-word
   match does not hold, the 3-word match still does!
 */

#include "siworin.h"
#include "siworin-displ.h"
#include "siworin-match.h"
#include "siworin-config.h"
#include "siworin-files.h"
#include <string.h>

static FILE *openHTMLfile (struct hfile_wordvector_s *matches);

static void DisplayContext (
   struct hfile_wordvector_s *matches,
   int num_matching_words, FILE *fph);

static void HighlightWords (
   struct hfile_wordvector_s *matches,
   int num_matching_words,
   FILE *fph,
   char *buf,
   int shown_context_before);

static void EditFilepathMakeURL (char *filepath);
static void ClearTags (char *buf);
static int  CalcWrdLen (char *wrd, FILE *fph, unsigned long offseth);
static void ClearInvalid (char *buf);

int siworin_displaymatches (
   struct hfile_wordvector_s *matches, int numelts_hfile_wordvector)
{
   FILE *fph;
   int i, num_matching_words;
   unsigned long prev_pathline = matches[0].vec[matches[0].currind].pathline;
   long prev_offseth  = matches[0].vec[matches[0].currind].offseth;
   int maxim_dista = atoi(siworin_config_value(SIWORIN_MAXIM_DISTA));

   for (i = 0;
        i < numelts_hfile_wordvector &&
        matches[i].currind < matches[i].numelts_vec &&
        matches[i].vec[matches[i].currind].pathline == prev_pathline &&
        (long)
        matches[i].vec[matches[i].currind].offseth - prev_offseth
           <= maxim_dista;
        i++)
      ;
   num_matching_words = i;

   #ifdef DEBUG2
   int j;
   fprintf(stdout, "<p>%s line %4d:\n", __FILE__, __LINE__);
   for (j = 0; j < numelts_hfile_wordvector; j++)
      fprintf(stdout, "<p>%d) %s\t%4lx %ld\n", j,
          matches[j].srch_word,
          matches[j].vec[matches[j].currind].offseth,
          matches[j].vec[matches[j].currind].offseth);
   #endif

   if (!(fph = openHTMLfile(matches)))
      SIWORIN_FAIL("finding the HTML file");

   DisplayContext(matches, num_matching_words, fph);
   fclose(fph);

   return 0;
}

#define FILEPATH_LEN  512
static FILE *openHTMLfile (struct hfile_wordvector_s *matches)
{
   FILE *fph;
   char *filepath;
   unsigned long offsetf;

   /* Find the path to the HTML file, indirectly via its index file. */
   if (fseek(siworin_fp_pathoffs,
         matches[0].vec[matches[0].currind].pathline
         * (SIWORIN_OFFSET_HEXDIGITS + 1),
         SEEK_SET) < 0)
      SIWORIN_FAIL("fseek");
   if (fscanf(siworin_fp_pathoffs, "%lx\n", &offsetf) != 1)
      SIWORIN_FAIL("fscanf");
   if (fseek(siworin_fp_paths, offsetf, SEEK_SET) < 0)
      SIWORIN_FAIL("fseek");
   if (!(filepath = malloc(FILEPATH_LEN)))
      SIWORIN_FAIL("malloc");
   if (fgets(filepath, FILEPATH_LEN, siworin_fp_paths) != filepath)
      SIWORIN_FAIL("fgets");

   /* Remove the trailing newline, and open the file. */
   if (filepath[0])
      filepath[strlen(filepath) - 1] = '\0';
   if (!(fph = fopen(filepath, "r")))
      SIWORIN_FAIL(filepath);

   /* Write the filename as a clickable link. */
   EditFilepathMakeURL(filepath);
   fprintf(stdout, "<p><a href=\"%s\" target=SearchResult>%s</a></p>\n",
      filepath, filepath);

   free(filepath);

   return fph;
}

static char *strong_befor = "<strong>";
static char *strong_after = "</strong>";

static void DisplayContext (
   struct hfile_wordvector_s *matches, int num_matching_words, FILE *fph)
{
   int show_context_befor =
      atoi(siworin_config_value(SIWORIN_SHOW_CONTEXT_BEFOR));
   /* Actual number, not necessarily as much as was wanted */
   int shown_context_befor;
   int show_context_after =
      atoi(siworin_config_value(SIWORIN_SHOW_CONTEXT_AFTER));

   int room_for_buffer;
   int bytes_to_read, bytes_read;
   char *buf;
   struct hfile_wordvector_s *fstmatch = matches;
   struct hfile_wordvector_s *lstmatch = matches + num_matching_words - 1;

   shown_context_befor = show_context_befor;

   if (fstmatch->vec[fstmatch->currind].offseth < (unsigned long)show_context_befor)
      shown_context_befor = fstmatch->vec[fstmatch->currind].offseth;

   bytes_to_read =
      shown_context_befor +
      (lstmatch->vec[lstmatch->currind].offseth -
       fstmatch->vec[fstmatch->currind].offseth) +
      strlen(lstmatch->srch_word) +
      show_context_after;

   room_for_buffer = bytes_to_read +
      num_matching_words *
         (strlen(strong_befor) + strlen(strong_after))
      +1; /* For nullbyte */

   if (!(buf = malloc(room_for_buffer)))
      SIWORIN_FAIL("malloc");
   if (fseek(fph,
         fstmatch->vec[fstmatch->currind].offseth - shown_context_befor,
         SEEK_SET) < 0)
      SIWORIN_FAIL("fseek");
   bytes_read = fread(buf, 1, bytes_to_read, fph);
   buf[bytes_read] = '\0';

   /* Note that tag clearing must not involve any shifts, because
      otherwise highlighting keywords by adding boldface won't work
      correctly later on. */
   ClearTags(buf);
   HighlightWords(
      matches, num_matching_words, fph, buf, shown_context_befor);
   ClearInvalid(buf);
   fprintf(stdout, "<p>%s</p>\n", buf);

   free(buf);
}

static void HighlightWords (
   struct hfile_wordvector_s *matches,
   int num_matching_words,
   FILE *fph,
   char *buf,
   int shown_context_before)
{
   struct {
      int strtpos;  /* Byte position where the word starts, in buf. */
      int nextpos;  /* Byte position just after the word. */
      int skipthis; /* Don't highlight for this word, was already done. */
   } *lentab = NULL;
   int i;

   if (!(lentab = malloc(num_matching_words * sizeof *lentab)))
      SIWORIN_FAIL("malloc");

   /* Build a table of word positions */
   for (i = 0; i < num_matching_words; i++)
   {
      lentab[i].strtpos = matches[i].vec[matches[i].currind].offseth -
                          matches[0].vec[matches[0].currind].offseth +
                          shown_context_before;
      lentab[i].nextpos = lentab[i].strtpos + CalcWrdLen(
         matches[i].srch_word,
         fph,
         matches[i].vec[matches[i].currind].offseth);

      /* If the same searchword was given more than once, the same
         position will occur multiple times. But highlighting should
         be done only once, for this position in the string. So mark
         this situation. */
      lentab[i].skipthis = 0;
      if (i > 0 && lentab[i].strtpos == lentab[i-1].strtpos)
         lentab[i].skipthis = 1;
   }

   /* Add HTML tags so the keywords will appear in bold.
      We do this from the end of the line backward, so
      the unedited positions remain valid, unaffected
      by shifts. */
   for (i = num_matching_words - 1; i >= 0; i--)
   {
      if (!lentab[i].skipthis)
      {
         memmove(buf + lentab[i].nextpos + strlen(strong_after),
                 buf + lentab[i].nextpos,
                 strlen(buf + lentab[i].nextpos) + 1);
         strncpy(buf + lentab[i].nextpos, strong_after, strlen(strong_after));

         memmove(buf + lentab[i].strtpos + strlen(strong_befor),
                 buf + lentab[i].strtpos,
                 strlen(buf + lentab[i].strtpos) + 1);
         strncpy(buf + lentab[i].strtpos, strong_befor, strlen(strong_befor));
      }
   }
   free(lentab);
}

static void EditFilepathMakeURL (char *filepath)
{
   char *repths = siworin_config_value(SIWORIN_FILEPATH_REPLACE_THIS);
   char *rep_by = siworin_config_value(SIWORIN_FILEPATH_REPLACE_BY  );
   int lenfpt;
   int lenths;
   int len_by;

   if (!repths || !rep_by)
      /* No replacement configured, so nothing to do. */
      return;

   /* Check if there enough room for the edit. Error if not,
      buffer should have been defined large enough. */
   lenfpt = strlen(filepath);
   lenths = strlen(repths);
   len_by = strlen(rep_by);
   if (lenfpt - lenths + len_by >= FILEPATH_LEN)
   {
      SIWORIN_FAIL("edit_filepath");
      return;
   }

   /* OK, everything is all right, now try to do the edit. */
   if (strncmp(repths, filepath, lenths) != 0)
      /* The string to be replaced is not there. So do nothing. */
      return;

   /* Now actually do the edit! */
   if (lenths != len_by)
      memmove(filepath + len_by, filepath + lenths, lenfpt + 1);
   strncpy(filepath, rep_by, len_by);
}

static void ClearTags (char *buf)
{
   char *cp;
   int incmnd = 0;

   if (atoi(siworin_config_value(SIWORIN_DONT_REMOVE_TAGS_CONTEXT)))
      return;

   for (cp = buf; *cp; cp++)
   {
      if (*cp == '<')
         incmnd = 1;

      if (!incmnd && *cp == '>')
         /* Incomplete tag at start of buffer, clear all. */
         memset(buf, ' ', cp - buf + 1);
      else if (incmnd && *cp != '>')
         /* Clear everything in the tag, including the first '<'. */
         *cp = ' ';
      else if (*cp == '>')
      {
         /* Clear also the last '>', after making a note that the
            tag has ended. */
         incmnd = 0;
         *cp = ' ';
      }
   }
}

/* See https://rudhar.com/sfreview/siworin/siworin08.htm#shy for an
   explanation why this is necessary.
   This implementation works well only for &shy; (and other entitities
   that happen to have this length), not for others like &ccedil; or
   &atilde; .

   Improvement 15 January 2023: we only count extra length for highlighting,
   if the found entities actually is "&shy;". Any entities weren't removed
   for the index, so then the length of the search word is already correct.
*/
static int CalcWrdLen (char *wrd, FILE *fph, unsigned long offseth)
{
   int len = strlen(wrd);
   char *shy = "&shy;";
   const int ls = strlen(shy);
   char *wbuf = NULL;
   int lookfrom = 0;
   char *fnd;
   int i;

   for (i = 0; /* Emergency break */ i < 10; i++)
   {
      if (!(wbuf = realloc(wbuf, len + ls + 1)))
         SIWORIN_FAIL("realloc");
      if (fseek(fph, offseth, SEEK_SET) < 0)
         SIWORIN_FAIL("fseek");
      memset(wbuf, '\0', len + ls + 1);
      fread(wbuf, 1, len + ls, fph);
      if ((fnd = strchr(wbuf + lookfrom, '&')) == NULL)
         break;
      /* Only count if actually &shy; */
      if (strncmp(fnd, shy, ls) == 0)
         len += ls;
      lookfrom = fnd - wbuf + 1;
   }
   free(wbuf);

   return len;
}

/* See https://rudhar.com/sfreview/siworin/siworin13.htm

   That was August 2021. Now in September I have meanwhile
   decided to do it in a much simpler way: replace any bytes
   at the start and end of the text buffer that are
   non-whitespace, by spaces or nulls respectively. That
   way, I not only clear invalid UTF-8 sequences, but also
   incomplete words.

   Of course this doesn't work for languages that do not
   use spaces between words, like Japanese, but then I
   never intended to support those in the first place.

   I leave the old function here, disabled. See below
   for the new one. */
#if 0
static void ClearInvalidChars (char *buf, int fromstart)
{
   char *cp;

   if (fromstart)
   {
      /* Blank out initial UTF-8 follow-up byte, that do not
         have their start byte included. Follow-ups start with
         the bits 10. */
      cp = buf;
      for (cp = buf; (*cp & 0xc0) == 0x80; cp++)
         *cp = ' ';
   }
   else
   {
      /* From the end, then. Remove all non-ASCII, because it
         may invalid. But not always. Then we sacrife one word. */
      cp = buf + strlen(buf);
      if (cp > buf) cp--;
      while ((*cp & 0x80) == 0x80)
      {
         *cp = '\0';
         if (cp > buf) cp--;
      }
   }
}
#endif
/* And this is the new function (note the slightly different function
   prototype): */
#include <ctype.h>
static void ClearInvalid (char *buf)
{
   char *cp, *ce;

   /* The checks on 50 bytes of distance are safety breaks,
      for cases of anomalous content. */
   for (cp = buf; *cp && !isspace(*cp) && cp < buf + 50; cp++)
      *cp = ' ';

   for (cp = ce = buf + strlen(buf) - 1;
        cp > buf && !isspace(*cp) && ce - cp < 50;
        cp--)
      *cp = '\0';
}
